/** @file
 * @brief PadicoTM components definition and API
 * @note shouldn't be included directly: #include <Padico/Puk.h> instead.
 * @ingroup Puk
 */

#ifndef PUK_COMPONENTS_H
#define PUK_COMPONENTS_H


/** @defgroup PadicoComponents API: PadicoTM components description
 * @ingroup Puk
 */

/** @defgroup PadicoComponentRegistry API: PadicoTM components registry- registration and query
 * @ingroup Puk
 */

/** @defgroup PadicoAssembly API: PadicoTM components assembly and instanciation interface
 * @see @ref PadicoComponents
 * @ingroup Puk
 */

/** @defgroup PadicoAdapter List of PadicoTM adapters
 */

/* ********************************************************* */
/* *** Components description ****************************** */


/** @addtogroup PadicoComponents
 * @{
 */

/* *** Types *********************************************** */

typedef struct puk_iface_s*puk_iface_t;
typedef struct puk_facet_s*puk_facet_t;
typedef struct puk_adapter_s*puk_component_t;
#define puk_adapter_t puk_component_t

/* *** Facets and attributes ******************************* */

/** An interface from a component
 */
struct puk_iface_s
{
  const char*name;
};
PUK_VECT_TYPE(puk_iface, struct puk_iface_s*)

/** Attach a driver to an iface (e.g. for iface 'foo'):
 *    -- create an iface lookup function
 *    -- create the receptacle type
 *    -- typed context and instance indirections
 * Prototypes:
 *   padico_iface_t padico_iface_foo(void);
 *   struct puk_receptacle_foo_s;
 *   void puk_instance_indirect_foo(puk_instance_t, struct puk_receptacle_foo_s*);
 *   void puk_context_indirect_foo(puk_instance_t, const char*label, struct puk_receptacle_foo_s*);
 *   const struct foo_driver_s*puk_adapter_get_driver_foo(puk_component_t, const char*label);
 */
#define PUK_IFACE_TYPE(IFACE_NAME, DRIVER_TYPE) \
  static inline puk_iface_t puk_iface_##IFACE_NAME(void) \
  { static puk_iface_t _iface = NULL; \
    if(!_iface) _iface = puk_iface_register(#IFACE_NAME); \
    return _iface; } \
  struct puk_receptacle_##IFACE_NAME##_s \
  { \
    DRIVER_TYPE*driver; \
    void*_status; \
  }; \
  static inline void puk_instance_indirect_##IFACE_NAME(puk_instance_t instance, const char*label, \
                                                        struct puk_receptacle_##IFACE_NAME##_s*receptacle) \
  { puk_instance_entry_point(instance, puk_iface_##IFACE_NAME(), label, &receptacle->driver, &receptacle->_status); } \
  static inline void puk_context_indirect_##IFACE_NAME(puk_instance_t instance, const char*label, \
                                                       struct puk_receptacle_##IFACE_NAME##_s*receptacle) \
  { puk_context_indirect(instance, puk_iface_##IFACE_NAME(), label, &receptacle->driver, &receptacle->_status); } \
  static inline const DRIVER_TYPE*puk_adapter_get_driver_##IFACE_NAME(puk_component_t adapter, const char*label) \
  { puk_facet_t f = puk_adapter_get_facet(adapter, puk_iface_##IFACE_NAME(), label); \
    const DRIVER_TYPE*d = (const DRIVER_TYPE*)(f?f->driver:NULL); \
    return d; }
									      
/** A facet from a component
 */
struct puk_facet_s
{
  puk_iface_t iface; /**< interface exposed by the facet */
  const char*label;  /**< label (name) of the facet */
  const void*driver; /**< implementation of the facet */
};
PUK_VECT_TYPE(puk_facet, struct puk_facet_s)

/** A receptacle from a component
 */
struct puk_uses_s
{
  puk_iface_t iface;
  const char*label;
};
PUK_VECT_TYPE(puk_uses, struct puk_uses_s)

/* *** Adapter ********************************************* */

/** An abstract adapter descriptor 
 */
struct puk_adapter_s
{
  const char*name;     /**< entry name (eg. "Circuit/SysIO") )*/
  puk_hashtable_t attrs; /**< attributes table: label -> value @note for undef attr: value=&attr_undefined */
  struct puk_uses_vect_s  uses;     /**< the interfaces needed by this adapter */
  struct puk_facet_vect_s provides; /**< the interfaces provided by this adapter */
};
PUK_VECT_TYPE(puk_adapter, puk_component_t)

puk_facet_t puk_adapter_get_facet(puk_component_t a, puk_iface_t iface, const char*label);

/** @} */

/* *** Register & query ************************************ */

/** @addtogroup PadicoComponentRegistry
 * @{
 */

/* ** interface registry */

puk_iface_t puk_iface_register(const char*name);
puk_iface_t puk_iface_lookup(const char*name);
void        puk_iface_destroy(puk_iface_t iface);

/* ** adapter registry */

puk_component_t puk_adapter_resolve(const char*adapter_name);

void puk_component_destroy(puk_component_t c);

/* ** registry locking handlers */
void padico_tm_comp_set_handlers(void(*_acquire)(void), void(*_release)(void));


/* *** Structure for ports declaration ********************* */

struct puk_component_port_s
{
  enum { PUK_COMPONENT_PROVIDES, PUK_COMPONENT_USES, PUK_COMPONENT_ATTR } kind;
  const char*label;
  union
  {
    struct
    {
      const char*iface_name;
      const void*driver;
    } provides;
    struct
    {
      const char*iface_name;
    } uses;
    struct
    {
      const char*value;
    } attr;
  } content;
};
#define puk_component_provides(IFACE_NAME, PORT_LABEL, DRIVER) \
  ((struct puk_component_port_s){ \
    .kind = PUK_COMPONENT_PROVIDES, \
    .label = (PORT_LABEL), \
    .content = {.provides = { .iface_name = (IFACE_NAME), .driver = (DRIVER)} }})

#define puk_component_uses(IFACE_NAME, PORT_LABEL) \
  ((struct puk_component_port_s){ \
    .kind = PUK_COMPONENT_USES, \
    .label = (PORT_LABEL), \
    .content = { .uses = { .iface_name = (IFACE_NAME) } } })
 
#define puk_component_attr(LABEL, VALUE) \
  ((struct puk_component_port_s){ \
    .kind = PUK_COMPONENT_ATTR, \
    .label = (LABEL), \
    .content = { .attr = { .value = (VALUE) } } })

puk_component_t puk_component_declare2(const char*adapter_name, int n_ports,
				       const struct puk_component_port_s ports[]);

#define puk_component_declare(COMPONENT_NAME, ...) \
  puk_component_declare2((COMPONENT_NAME), \
    sizeof((struct puk_component_port_s[]){ __VA_ARGS__ }) / \
     sizeof(struct puk_component_port_s), \
    ((struct puk_component_port_s[]){ __VA_ARGS__ }))

/* legacy puk_adapter_* interface */

#define puk_adapter_declare(ADAPTER_NAME, ...) \
  puk_component_declare(ADAPTER_NAME, ##__VA_ARGS__)

#define puk_adapter_provides(IFACE_NAME, DRIVER) \
  puk_component_provides((IFACE_NAME), (IFACE_NAME), (DRIVER))

#define puk_adapter_uses(IFACE_NAME) \
  puk_component_uses((IFACE_NAME), (IFACE_NAME))

#define puk_adapter_attr(LABEL, VALUE) \
  puk_component_attr(LABEL, VALUE)

/** @} */


/* ********************************************************* */
/* *** Assembly ******************************************** */

/** @addtogroup PadicoAssembly
 * @{
 */

/* *** Types ********************************************* */

typedef struct puk_component_conn_s*puk_component_conn_t;
typedef struct puk_generic_context_s*puk_context_t;
typedef struct puk_instance_s*puk_instance_t;

/* *** Adapter context *********************************** */

/** a component port connection, between a receptacle and a facet */
struct puk_component_conn_s
{
  struct puk_uses_s*receptacle;
  struct puk_facet_s*facet;
  puk_context_t context;
};
PUK_VECT_TYPE(puk_component_conn, struct puk_component_conn_s)


/** a context for an adapter in an assembly
 * (connections, attributes)
 */
struct puk_generic_context_s
{
  const char*         id; /**< context ID, as given in the XML assembly description */
  puk_component_t  adapter; /**< the adapter this context corresponds to */
  puk_component_t assembly; /**< the assembly this context belongs to */
  struct puk_component_conn_vect_s conns; /**< interfaces connections */
  puk_hashtable_t  attrs; /**< attributes */
};
PUK_VECT_TYPE(puk_context, puk_context_t)

const char*puk_context_getattr(puk_context_t, const char*);

/* *** Composite ******************************************* */

/** an assembly made of several adapters and their connections */
struct puk_composite_content_s
{
  struct puk_composite_driver_s
  {
    struct puk_composite_content_s*content;
  } driver;
  struct puk_component_conn_vect_s entry_points; /**< connections for the entry points */
  puk_hashtable_t contexts; /**< contexts in this assembly- hashed by context ID */
  int checked; /**< whether the assembly as been checked (ok to be instanciated) */
};


padico_string_t puk_adapter_serialize(puk_component_t);

puk_component_t puk_adapter_parse(const char*);

puk_component_t puk_adapter_parse_file(const char*filename);

void puk_adapter_check(puk_component_t assembly);


/* ********************************************************* */
/* *** Instanciation *************************************** */


/* *** Assembly instance ********************************* */

/** An instance of an adapter */
struct puk_instance_s
{
  puk_component_t adapter; /**< the adapter of which this is an instance */
  void*status;           /**< adapter-specific status */
  puk_context_t context; /**< context of the adapter (if any) */
  puk_instance_t container;
};

/** Status of a composite component */
struct puk_composite_status_s
{
  puk_hashtable_t instance_table; /**< hash: context -> instance */
  puk_instance_t instance;        /**< component instance containing this composite status */
};

/** a predefined descriptor for the instanciation interface: PadicoAdapter */
struct puk_adapter_driver_s
{
  void*(*instanciate)(puk_instance_t, puk_context_t);
  void (*destroy)(void*);
};

/** Creates a instance of a given component. */
puk_instance_t puk_component_instanciate(puk_component_t adapter, puk_instance_t container, puk_context_t context);

#define puk_adapter_instanciate(adapter) puk_component_instanciate(adapter, NULL, NULL)

/** Destroys an instance */
void puk_instance_destroy(puk_instance_t instance);


/** Finds driver & status of a sub-context */
void puk_context_indirect(puk_instance_t instance,
			  puk_iface_t iface,
			  const char*label,
			  void*_driver, void *_status);

/** Finds an entry point of an instance */
void puk_instance_entry_point(puk_instance_t instance,
			      puk_iface_t iface,
			      const char*label,
			      void*_driver, void*_status);

/** finds in 'container' an instance with the same context as 'local' */
puk_instance_t puk_instance_find_context(puk_instance_t container, puk_instance_t local);

/** Get attributes for an instance */
static inline const char*puk_instance_getattr(puk_instance_t instance, const char*label)
{
  const char*attr = (instance->context) ? 
    puk_context_getattr(instance->context, label) :
    (instance->adapter->attrs ? 
     (const char*)puk_hashtable_lookup(instance->adapter->attrs, label) : NULL);
  return attr;
}

/* *** various inline functions */

PUK_IFACE_TYPE(PadicoAdapter, struct puk_adapter_driver_s)
PUK_IFACE_TYPE(PadicoComposite, struct puk_composite_driver_s)

/** Finds adapter instance status for a given (in-context) instance */
static inline void*puk_instance_self_status(puk_instance_t instance)
{
  return instance->status;
}

/** returns the root container (outermost containing instance) for the given instance.
 */
static inline puk_instance_t puk_instance_get_root_container(puk_instance_t instance)
{
  while(instance->container != NULL)
    instance = instance->container;
  return instance;
}


/** @} */

/* *** Enumerator **************************************** */

/* Adapter enumerators provide an _ordered_ enumeration of 
 * the contexts of an assembly, so as to be able to calculate
 * inter-adapter dependancies when instanciating assemblies.
 */

typedef struct puk_adapter_enumerator_s*puk_adapter_enumerator_t;

/** direction of adapter enumeration : top-down, or bottom-up
 */
enum adapter_enumerator_way_e
  { 
    TOP_DOWN = 1,
    BOTTOM_UP
  };

puk_adapter_enumerator_t puk_adapter_enumerator_new(puk_component_t assembly, int way);
puk_context_t            puk_adapter_enumerator_next(puk_adapter_enumerator_t e);
void                     puk_adapter_enumerator_delete(puk_adapter_enumerator_t e);


/* ********************************************************* */



/* *** Adapter-specific data ******************************* */
/* *** (obsolete) */ 

/** @addtogroup PadicoComponents
 * @{
 */

struct puk_adapter_specific_bucket_s
{
  puk_component_t adapter;
  void*content;
};
PUK_VECT_TYPE(puk_adapter_specific_bucket, struct puk_adapter_specific_bucket_s)

typedef struct puk_adapter_specific_bucket_vect_s puk_adapter_specific_t;

#ifndef __cplusplus
static inline void puk_adapter_specific_init(puk_adapter_specific_t*as)
  __attribute__((deprecated));
static inline void puk_adapter_specific_destroy(puk_adapter_specific_t*as)
  __attribute__((deprecated));
static inline void*puk_adapter_getspecific(puk_adapter_specific_t*as,
						 puk_component_t a)
  __attribute__((deprecated));
static inline void puk_adapter_setspecific(puk_adapter_specific_t*as,
						 puk_component_t a,
						 void*value)
  __attribute__((deprecated));
     
static inline void puk_adapter_specific_init(puk_adapter_specific_t*as)
{
  puk_adapter_specific_bucket_vect_init(as);
}
static inline void puk_adapter_specific_destroy(puk_adapter_specific_t*as)
{
  puk_adapter_specific_bucket_vect_destroy(as);
}
static inline void*puk_adapter_getspecific(puk_adapter_specific_t*as,
						 puk_component_t a)
{
  puk_adapter_specific_bucket_vect_itor_t i;
  puk_vect_foreach(i, puk_adapter_specific_bucket, as)
    {
      if(i->adapter == a)
	return i->content;
    }
  return NULL;
}
static inline void puk_adapter_setspecific(puk_adapter_specific_t*as,
						 puk_component_t a,
						 void*value)
{
  puk_adapter_specific_bucket_vect_itor_t i;
  puk_vect_foreach(i, puk_adapter_specific_bucket, as)
    {
      if(i->adapter == a)
	{
	  i->content = value;
	  return;
	}
    }
  struct puk_adapter_specific_bucket_s b = { .adapter = a, .content = value };
  puk_adapter_specific_bucket_vect_push_back(as, b);
}
#endif /* __cplusplus */

/** @} */


#endif /* PUK_COMPONENTS_H */

