/** @file
 ** @brief   Public API header for Puk-ABI.
 *  @ingroup PukABI
 *  @author  Alexandre Denis
 *  @note    Only stuff defined in *this file* is part of the API, even
 *           if internals are #include'd for inlined functions.
 *  @warning Do not include this file in PukABI itself! MKSYMxX macros 
 *           have different declerations.
 */


/** @defgroup PukABI API: Puk ABI -- dynamic ABI manager
 * @ingroup Puk
 */

#ifndef PUK_ABI_H
#define PUK_ABI_H

#define PUK_ABI

/* include internals for inlined and #define'd functions */
#include "Puk-ABI-internals.h"
#include <sys/types.h>

extern void puk_abi_init(void);

/* ********************************************************* */

/** @internal
 * @brief Generates symbol types. 
 */
#define MKSYMFX(SYM, RETTYPE, ARGSDECL, ARGSUSE, PRE, POST,  ISREAL)     PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL)
#define MKSYMVX1(SYM, RETTYPE, ARGSDECL, ARG1, PROCESSING, ISREAL)       PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL)
#define MKSYMVX2(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, PROCESSING, ISREAL) PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL)
#define MKSYMVX3(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, ARG3, PROCESSING, ISREAL) PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL)
#define MKSYMPX(SYM, RETTYPE, ARGSDECL, ARGSUSE, PRE, POST,  ISREAL)    PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL)

#define MKSYMNOSUPPORT(SYM)   PUK_ABI_MKTYPE(SYM, void, (void))

#include "Puk-ABI-symbols.h"

/* ********************************************************* */


/* *** Methods to access symbols *************************** */

/** Call the *real* libc symbol (always REAL, no locking)- it is up to the 
 * user to appropriately puk_spinlock_{acquire|release}().
 */
#define PUK_ABI_REAL(SYM)    (*((_PASTE(SYM, _func_t))(PUK_SYM_REAL(SYM))))

/** Call the *virtual* symbol (always virtual).
 * @note not really usefull, except for completeness of PukABI API.
 */
#define PUK_ABI_VIRTUAL(SYM) (*((_PASTE(SYM, _func_t))(PUK_SYM_VIRTUAL(SYM))))

/** Call the *stub* symbol (auto choice) 
 */
#define PUK_ABI_STUB(SYM)   (SYM) /* PUK_SYM_STUB(SYM) */

/** Call the *wrap* symbol (real + puk_spinlock_{acquire|release}() ) 
 */
#define PUK_ABI_WRAP(SYM)   PUK_SYM_WRAP(SYM)


/* *** errno management ************************************ */

#define puk_abi_seterrno(VAL)    ( (*(puk_abi_errno_stub())) = (VAL) )
#define puk_abi_geterrno()       (*(puk_abi_errno_stub()))
#define __puk_abi_wrap__errno    (*(puk_abi_errno_stub()))
#define puk_abi_real_errno()     (*(puk_abi_errno_real()))

/* *** memory management *********************************** */

typedef void*(*puk_mem_register_t)  (const void*ptr, size_t len);
typedef void (*puk_mem_unregister_t)(const void*ptr, void*key);
void puk_mem_set_handlers(puk_mem_register_t reg, puk_mem_unregister_t unreg);
void*puk_mem_reg(const void*ptr, size_t len);
void puk_mem_unreg(const void*ptr);

/* *** FD management *************************************** */

extern int  puk_abi_vfd_alloc(void);
extern void puk_abi_vfd_free(int abi_fd);

/* *** Various handlers
 */
/** sets the handler for a virtual symbol */
#define puk_abi_set_virtual(SYM, FUNC) \
{ \
  PUK_SYM_DESCNAME(SYM).ptr_virtual = (FUNC); \
}

/** sets the main spinlock handler */
extern void puk_abi_set_spinlock_handlers(void (*acquire)(void),
					  void (*release)(void),
					  int (*trylock)(void));
/** sets the errno location function */
extern void puk_abi_set_errno_handler(int*(*errno_handler)(void));


#endif /* PUK_ABI_H */
