/** @file 
 * @brief Puk memory management
 * @ingroup Puk
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "Puk-ABI.h"

static void*puk_mem_alloc(unsigned long);
static void puk_mem_free(void*);
static void*puk_mem_realloc(void*ptr, unsigned long size);
static void*puk_mem_calloc(size_t nelem, size_t elsize);
static char*puk_mem_strdup(const char*s);
static void(*puk_mem_invalidate_hook)(void*ptr) = NULL;

/* ********************************************************* */

void puk_abi_mem_init()
{
  puk_abi_set_virtual(malloc,  &puk_mem_alloc);
  puk_abi_set_virtual(free,    &puk_mem_free);
  puk_abi_set_virtual(realloc, &puk_mem_realloc);
  puk_abi_set_virtual(calloc,  &puk_mem_calloc);
  puk_abi_set_virtual(strdup,  &puk_mem_strdup);
}

/* ********************************************************* */

#define PUK_MEM_CACHE_SIZE 8

static struct
{
  struct puk_mem_reg_s
  {
    const void*ptr;
    size_t len;
    void*key;
    int refcount;
  } cache[PUK_MEM_CACHE_SIZE];
  puk_mem_register_t   reg;
  puk_mem_unregister_t unreg;
} puk_mem =
  {
    .reg   = NULL,
    .unreg = NULL,
    .cache = { { .refcount = 0, .ptr = NULL, .len = 0 } }
  };


static void puk_mem_cache_flush(struct puk_mem_reg_s*r)
{
  (*puk_mem.unreg)(r->ptr, r->key);
  r->ptr = NULL;
  r->len = 0;
  r->key = 0;
}
static struct puk_mem_reg_s*puk_mem_cache_lookup(const void*ptr)
{
  assert(ptr != NULL);
  int i;
  for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
    {
      struct puk_mem_reg_s*r = &puk_mem.cache[i];
      if(ptr >= r->ptr && (ptr < r->ptr + r->len))
	return r;
    }
  return NULL;
}
static struct puk_mem_reg_s*puk_mem_slot_lookup(void)
{
  static unsigned int victim = 0;
  int i;
  for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
    {
      struct puk_mem_reg_s*r = &puk_mem.cache[i];
      if(r->ptr == NULL && r->len == 0)
	{
	  return r;
	}
    }
  victim++;
  struct puk_mem_reg_s*r = &puk_mem.cache[victim % PUK_MEM_CACHE_SIZE];
  while(r->refcount > 0)
    {
      victim++;
      r = &puk_mem.cache[victim % PUK_MEM_CACHE_SIZE];
    }
  (*puk_mem.unreg)(r->ptr, r->key);
  r->ptr = NULL;
  r->len = 0;
  r->key = NULL;
  return r;
}

static void puk_mem_invalidate(void*ptr)
{
  if(ptr)
    {
      struct puk_mem_reg_s*r = puk_mem_cache_lookup(ptr);
      while(r)
	{
	  puk_mem_cache_flush(r);
	  r = puk_mem_cache_lookup(ptr);
	}
    }
}

/* ********************************************************* */

void puk_mem_set_handlers(puk_mem_register_t reg, puk_mem_unregister_t unreg)
{
  puk_spinlock_acquire();
  puk_mem.reg = reg;
  puk_mem.unreg = unreg;
  puk_mem_invalidate_hook = &puk_mem_invalidate;
  puk_spinlock_release();
}
void puk_mem_unreg(const void*ptr)
{
  assert(ptr != NULL);
  puk_spinlock_acquire();
  struct puk_mem_reg_s*r = puk_mem_cache_lookup(ptr);
  if(r)
    {
      r->refcount--;
    }
  puk_spinlock_release();
}
void*puk_mem_reg(const void*ptr, size_t len)
{
  assert(ptr != NULL);
  puk_spinlock_acquire();
  struct puk_mem_reg_s*r = puk_mem_cache_lookup(ptr);
  if(r && (ptr + len <= r->ptr + r->len))
    {
      r->refcount++;
    }
  else
    {
      if(r)
	{
	  puk_mem_cache_flush(r);
	}
      r = puk_mem_slot_lookup();
      r->ptr = ptr;
      r->len = len;
      r->refcount = 1;
      r->key = (*puk_mem.reg)(ptr, len);
    }
  puk_spinlock_release();
  return r->key;
}

/* ********************************************************* */

static void*puk_mem_alloc(unsigned long len)
{
  puk_spinlock_acquire();
  void *tmp = PUK_ABI_REAL(malloc)(len);
  puk_spinlock_release();
  return tmp;
}

static void puk_mem_free(void*ptr)
{
  puk_spinlock_acquire();
  if(puk_mem_invalidate_hook)
    (*puk_mem_invalidate_hook)(ptr);
  PUK_ABI_REAL(free)(ptr);
  puk_spinlock_release();
}

static void*puk_mem_realloc(void*ptr, unsigned long size)
{
  puk_spinlock_acquire();
  if(puk_mem_invalidate_hook)
    (*puk_mem_invalidate_hook)(ptr);
  void*rc = PUK_ABI_REAL(realloc)(ptr, size);
  puk_spinlock_release();
  return rc;
}

static void*puk_mem_calloc(size_t nelem, size_t elsize)
{
  void*ptr = puk_mem_alloc(elsize*nelem);
  memset(ptr, 0, nelem*elsize);
  return ptr;
}

static char*puk_mem_strdup(const char*s1)
{
  char*p = NULL;
  if(s1)
    {
      const int size = strlen(s1) + 1;
      p = puk_mem_alloc(size);
      if(!p)
	return NULL;
      memcpy(p, s1, size);
    }
  return p;
}

