/** @file
 * @ingroup Puk
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <string.h>
#include <poll.h>
#include <fcntl.h>

#include "Puk-ABI.h"

#define PUK_ABI_SYMINIT(SYM, PTR) ((PUK_SYM_DESCNAME(SYM)).ptr_real = (PTR))

static const void*puk_sym_resolve_next(const char*symbol);


void puk_abi_preinit(void) __attribute__((constructor));
void puk_abi_preinit(void)
{
#ifdef PADICO_ENABLE_OUTPUT
  fprintf(stderr, "PukABI: preinit.\n");
#endif

  /* ** pre-init symbol table with bare minimum for dlsym() */
  /* pre-load malloc/free */
  void*tmp_malloc = malloc(1);
  free(tmp_malloc);
  /* resolve errno */
  const void*next_errno = puk_sym_resolve_next(PUK_SYM_DESCNAME(puk_abi_errno_sym).name);
  PUK_ABI_SYMINIT(puk_abi_errno_sym, next_errno);
  /* hook memory wrappers */
  puk_abi_mem_init();
}


#define MAXMSG 1024
static ssize_t (*write_func)(int fildes, const void *buf, size_t nbyte) = NULL;

#ifdef PUK_ABI_TRACE_ON
static int puk_sym_debug = 1;
#else
static int puk_sym_debug = 0;
#endif

#ifdef PUK_ABI_CHECK_ASSERT
extern void puk_abi_do_assert(int cond, const char*file, int line, const char*func)
{
  if(!cond)
    {
      puk_abi_fatal("PukABI: assertion failed in %s %s:%d\n", func, file, line);
    }
}
#endif

/** resolve next entry in symbol table (supposedly 'real' symbol from libc)
 */
static const void*puk_sym_resolve_next(const char*symbol)
{
  const void*fun = dlsym(RTLD_NEXT, symbol);
  if(fun == NULL)
    {
      puk_abi_trace("##### WARNING: symbol %s not found!", symbol);
    }
  return fun;
}

/** resolve next symbol and registers it as 'real' in PukABI table
 */
extern const void*puk_abi_real_symbol(struct puk_symbol_s*sym)
{
  const void*fun = puk_sym_resolve_next(sym->name);
  if(!fun)
    {
      puk_abi_fatal("Puk-ABI: REAL symbol %s not found.\n", sym->name);
    }
  puk_abi_trace("PukABI: resolving REAL symbol '%s'.", sym->name);
  sym->ptr_real = fun;
  return fun;
}

void puk_abi_trace(char*msg, ...)
{
#ifdef PUK_ABI_SHOW_STACKTOP
  int stack_top = 0;
#endif
  char debugmsg[MAXMSG];
  if(puk_sym_debug)
    {
      int s = 0;
      int done = 0;
      va_list ap;
      va_start(ap, msg);
#ifdef PUK_ABI_TRACE_PID
      s = snprintf(debugmsg, MAXMSG, "Puk-ABI: [%d] ", getpid());
#endif
#ifdef PUK_ABI_SHOW_STACKTOP
      s = snprintf(debugmsg, MAXMSG, "%lX: ", ((unsigned long)&stack_top)>>12);
#endif
      s += vsnprintf(debugmsg + s, MAXMSG, msg, ap);
      s += snprintf(debugmsg + s, MAXMSG, "\n");
      if(!write_func)
	{
	  write_func = puk_sym_resolve_next("write");
	}
      do
	{
	  done += (*write_func)(2, debugmsg, s-done);
	}
      while(done < s);
      va_end(ap);
    }
}

void puk_abi_print(char*msg, ...)
{
#ifdef PUK_ABI_SHOW_STACKTOP
  int stack_top = 0;
#endif
  char debugmsg[MAXMSG];
  int s = 0;
  int done = 0;
  va_list ap;
  va_start(ap, msg);
  s += vsnprintf(debugmsg + s, MAXMSG, msg, ap);
#ifdef PUK_ABI_SHOW_STACKTOP
  s +=  snprintf(debugmsg + s, MAXMSG, " (stack top: %p)", (void*)(((unsigned long)&stack_top)>>12));
#endif
  s +=  snprintf(debugmsg + s, MAXMSG, "\n");
  if(!write_func)
    {
      write_func = puk_sym_resolve_next("write");
    }
  do
    {
      done += (*write_func)(2, debugmsg, s-done);
    }
  while(done < s);
  va_end(ap);
}

void puk_abi_fatal(char*msg, ...)
{
  char debugmsg[MAXMSG];
  int s;
  va_list ap;
  va_start(ap, msg);
  s = vsnprintf(debugmsg, MAXMSG, msg, ap);
  if(!write_func)
    {
      write_func = puk_sym_resolve_next("write");
    }
  (*write_func)(2, debugmsg, s);
  va_end(ap);
  abort();
}

/* ********************************************************* */

extern void puk_abi_set_spinlock_handlers(void (*acquire)(void),
					  void (*release)(void),
					  int (*trylock)(void))
{
  puk_spinlock.acquire = acquire;
  puk_spinlock.release = release;
  puk_spinlock.trylock = trylock;
}

static void puk_abi_disable_preload(void)
{
  putenv("LD_PRELOAD=");
}

void puk_abi_init(void)
{
  puk_abi_disable_preload();
  return;
}


