/** @file
 * @ingroup PukABI
 */

/** @note No other include file is allowed! */
#define PUK_ABI_LIBC_HEADERS_FORBIDDEN
#include "Puk-ABI-internals.h"

#include <stdarg.h>

/* Puk-ABI management objects
 */

static void default_spinlock_acquire(void) { /* do nothing */ }
static void default_spinlock_release(void) { /* do nothing */ }
static int  default_spinlock_trylock(void) { return 0; }

struct puk_spinlock_s puk_spinlock = 
  { 
    .acquire = &default_spinlock_acquire,
    .release = &default_spinlock_release,
    .trylock = &default_spinlock_trylock
  };

#ifdef PADICO_ENABLE_PUKABI_FSYS

/* File descriptor virtualization.  */

struct puk_fd_s
{
  int kind; /* 0: empty; 1: real; 2: virtual */
  int fd;
};
#define PUK_FDS_NONE 0
#define PUK_FDS_REAL 1
#define PUK_FDS_VFD  2

#if defined(PUK_ABI_SHOW_RECOVERY) || defined(PUK_ABI_TRACE_ON)
static const char*puk_fds_kind_label[] =
  {
    [ PUK_FDS_NONE ] = "none",
    [ PUK_FDS_REAL ] = "real",
    [ PUK_FDS_VFD  ] = "vfd"
  };
#endif

/* index = ABI fd; content = phys fd / vfd */
static struct puk_fd_s puk_abi_fds[PUK_ABI_FD_MAX] = 
  {
    [0] = { .kind = PUK_FDS_REAL, .fd = 0 }, /* stdin  */
    [1] = { .kind = PUK_FDS_REAL, .fd = 1 }, /* stdout */
    [2] = { .kind = PUK_FDS_REAL, .fd = 2 }, /* stderr */
    [3 ... PUK_ABI_FD_MAX-1] = { .kind = PUK_FDS_NONE, .fd = -1 } /* set remaining entries to NONE */
  };

/* index: hfd/vfd; content: abi fd */
static int puk_abi_reverse_fds[2][PUK_ABI_FD_MAX] = 
  {
    { /* reverse mapping for kind=real */
      [0] = 0,
      [1] = 1,
      [2] = 2,
      [3 ... PUK_ABI_FD_MAX-1] = -1 
    }, 
    { /* reverse mapping for kind=vfd  */
      [0 ... PUK_ABI_FD_MAX-1] = -1
    } 
  };

/* ********************************************************* */

static inline void puk_fd_register(int abi_fd, int real_fd, int kind)
{
  puk_abi_assert(abi_fd >= 0 && abi_fd < PUK_ABI_FD_MAX);
  puk_abi_assert(real_fd >= 0 && real_fd < PUK_ABI_FD_MAX);
  puk_abi_assert(kind == PUK_FDS_REAL || kind == PUK_FDS_VFD);
  puk_abi_fds[abi_fd] = (struct puk_fd_s){ .kind = kind, .fd = real_fd };
  puk_abi_reverse_fds[kind-1][real_fd] = abi_fd;
}

static inline void puk_fd_unregister(int abi_fd)
{
  puk_abi_assert(abi_fd >= 0 && abi_fd < PUK_ABI_FD_MAX);
  puk_abi_assert(puk_abi_fds[abi_fd].fd >= 0 && puk_abi_fds[abi_fd].fd < PUK_ABI_FD_MAX);
  puk_abi_reverse_fds[puk_abi_fds[abi_fd].kind-1][puk_abi_fds[abi_fd].fd] = -1;
  puk_abi_fds[abi_fd].kind = PUK_FDS_NONE;
}

static inline int puk_fd_to_abi(int real_fd, int kind)
{
  int abi_fd = -1;
  puk_abi_assert(real_fd >= 0 && real_fd < PUK_ABI_FD_MAX);
  abi_fd = puk_abi_reverse_fds[kind-1][real_fd];
  if(abi_fd == -1)
    {
#ifdef PUK_ABI_SHOW_RECOVERY
      puk_abi_print("PukABI: warning- fd=%d; kind=%s- unknown descriptor. Automatic recovery.",
		    real_fd, puk_fds_kind_label[kind]);
#endif
      puk_fd_register(real_fd, real_fd, kind);
      abi_fd = real_fd;
    }
  return abi_fd;
}

static inline void puk_fd_from_abi(int abi_fd, int*real_fd, int*kind)
{
#ifdef PUK_ABI_CHECK_ASSERT
  puk_abi_assert(real_fd != NULL && kind != NULL);
#endif
  if(abi_fd < 0)
    {
#ifdef PUK_ABI_SHOW_RECOVERY
      puk_abi_print("PukABI: warning- fd=%d is not a valid ABI fd- cannot translate.", abi_fd);
#endif
      *real_fd = -1;
      *kind = PUK_FDS_REAL;
      return; /** @note This should be an 'abort()'. However, libresolv performs
	       * 'read()' on fd=-1 on purpose. Give a sensible answer to make
	       * libresolv happy :-/
	       */
    }
#ifdef PUK_ABI_CHECK_ASSERT
  puk_abi_assert(abi_fd >= 0 && abi_fd < PUK_ABI_FD_MAX);
#endif
  if(abi_fd < PUK_ABI_FD_OFFSET && puk_abi_fds[abi_fd].kind == PUK_FDS_NONE)
    {
#ifdef PUK_ABI_RECOVER_BADFD
      /* defaults to _real_ */
#ifdef PUK_ABI_SHOW_RECOVERY
      puk_abi_print("PukABI: warning- fd=%d is not allocated. Automatic recovery as *real*.", abi_fd);
#endif
      *real_fd = abi_fd;
      *kind    = PUK_FDS_REAL;
      puk_fd_register(abi_fd, abi_fd, PUK_FDS_REAL);
#else
      puk_fds_dump();
      puk_sym_fatal("Bad FD: %d\n", abi_fd);
#endif
    }
  else
    {
#ifdef PUK_ABI_CHECK_ASSERT
      puk_abi_assert(puk_abi_fds[abi_fd].kind != PUK_FDS_NONE);
#endif
      *real_fd = puk_abi_fds[abi_fd].fd;
      *kind    = puk_abi_fds[abi_fd].kind;
    }
}

/* ********************************************************* */

void puk_fds_dump(void)
{
  int i;
  for(i=0; i<PUK_ABI_FD_MAX; i++)
    {
      switch(puk_abi_fds[i].kind)
	{
	case 1:
	  puk_abi_print("FD abi:%d -> real:%d", i, puk_abi_fds[i].fd);
	  break;
	case 2:
	  puk_abi_print("FD abi:%d -> virtual:%d", i, puk_abi_fds[i].fd);
	  break;
	}
    }
}

extern int puk_abi_vfd_alloc(void)
{
  int abi_fd = PUK_ABI_FD_OFFSET;
  puk_abi_trace("PukABI: vfd_alloc.");
  while(puk_abi_fds[abi_fd].kind != 0 && abi_fd < PUK_ABI_FD_MAX)
    abi_fd++;
  if(abi_fd >= PUK_ABI_FD_MAX)
    {
      puk_abi_print("Puk-ABI: out of file descriptors (MAX=%d)\n", PUK_ABI_FD_MAX);
      return -1;
    }
  puk_fd_register(abi_fd, abi_fd, PUK_FDS_VFD);
  puk_abi_trace("adding FD abi:%d -> virtual:%d", abi_fd, abi_fd);
  return abi_fd;
}

extern void puk_abi_vfd_free(int abi_fd)
{
  puk_abi_trace("PukABI: deleting FD abi:%d -> virtual:%d", abi_fd, puk_abi_fds[abi_fd].fd);
  puk_fd_unregister(abi_fd);
}

static inline int puk_fds_alloc(int kind, int fd, const char*caller)
{
  int abi_fd = PUK_ABI_FD_OFFSET;
  puk_abi_trace("PukABI: puk_fds_alloc()- kind=%d; fd=%d", kind, fd);
  while(puk_abi_fds[abi_fd].kind != 0 && abi_fd < PUK_ABI_FD_MAX)
    abi_fd++;
  if(abi_fd >= PUK_ABI_FD_MAX)
    puk_abi_fatal("Puk-ABI: out of file descriptors (MAX=%d)\n", PUK_ABI_FD_MAX);
  puk_fd_register(abi_fd, fd, kind);
  puk_abi_trace("PukABI: adding FD abi:%d -> real:%d (%s)", abi_fd, fd, caller);
  return abi_fd;
}
static inline void puk_fds_free(int abi_fd)
{
  puk_abi_trace("PukABI: deleting FD abi:%d -> real:%d", abi_fd, puk_abi_fds[abi_fd].fd);
  puk_fd_unregister(abi_fd);
}

extern int puk_fds_from_abi(int abi_fd)
{
  int real_fd = -1, kind = PUK_FDS_NONE;
  puk_fd_from_abi(abi_fd, &real_fd, &kind);
  return real_fd;
}

extern int puk_fds_isreal(int abi_fd)
{
  int real_fd = -1, kind = PUK_FDS_NONE;
#ifdef PUK_ABI_TRACE_ON
  puk_abi_trace("PukABI: checking whether FD abi:%d is real (kind=%s).",
		abi_fd, puk_fds_kind_label[puk_abi_fds[abi_fd].kind]);
#endif
  puk_fd_from_abi(abi_fd, &real_fd, &kind);
  return (kind == PUK_FDS_REAL);
}

extern int puk_fds_lookup(int real_fd, int kind)
{
  return puk_fd_to_abi(real_fd, kind);
}

#endif /* PADICO_ENABLE_PUKABI_FSYS */

/* ********************************************************* */
/* ** Stub and wrapper builders **************************** */

/* ** __ Aliasing rules for PukABI __
 *
 * 'foo'            is an alias for '__puk_abi_foo'
 * 'foo_func_t'     is the type of function 'foo'
 * '__puk_abi_desc__foo__' is a "struct puk_symbol_s" describing symbol 'foo'
 * '__puk_abi_stub__foo__' is a stub with automatic selection for real/virtual
 * '__puk_abi_wrap__foo' is a wrapper for real+lock
 */

#define _MK_ALIAS(NAME, ALIAS) \
  extern __typeof (NAME) ALIAS __attribute__ ((alias (_TOSTRING(NAME))));

/** Generates symbol stub-
 *  function version 
 * OPTION: bitmap-
 * OPTION & 0x01: use_real
 * OPTION & 0x02: use_vararg
 */
#define _PUKABI_MK_SYM_STUBF(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, OPTIONS) \
  RETTYPE PUK_SYM_STUB(SYM) ARGSDECL { \
    RETTYPE result; \
    const int _options = (OPTIONS); \
    const int _will_use_real = _options; \
    if(_will_use_real){ \
      const SYM##_func_t fun = (SYM##_func_t)(PUK_SYM_REAL(SYM)); \
      puk_spinlock_acquire(); \
      PREPROCESSING \
      result = ((*fun) ARGSUSE); \
      PUK_ABI_ERRNO_INHERIT; \
      POSTPROCESSING \
      puk_spinlock_release(); \
    } else { \
      const SYM##_func_t fun = (SYM##_func_t)(PUK_SYM_VIRTUAL(SYM)); \
      PREPROCESSING \
      result = ((*fun) ARGSUSE); \
      POSTPROCESSING \
    } \
    return result; }

#define _PUKABI_MK_SYM_WRAPF(SYM, RETTYPE, ARGSDECL, ARGSUSE) \
  RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL { \
    RETTYPE result; \
    const SYM ## _func_t fun = (SYM ## _func_t)(PUK_SYM_REAL(SYM)); \
    puk_spinlock_acquire(); \
    result = ((*fun) ARGSUSE); \
    PUK_ABI_ERRNO_INHERIT; \
    puk_spinlock_release(); \
    return result; } 

#define _PUKABI_MK_SYM_STUBV(SYM, RETTYPE, ARGSDECL, PROCESSING, OPTIONS) \
  RETTYPE PUK_SYM_STUB(SYM) ARGSDECL { \
    RETTYPE _result; \
    const int _options = (OPTIONS); \
    const int _will_use_real = _options; \
    if(_will_use_real){ \
      const SYM##_func_t PUK_UNUSED _fun = (SYM##_func_t)(PUK_SYM_REAL(SYM)); \
      puk_spinlock_acquire(); \
      PUK_ABI_ERRNO_INHERIT; \
      PROCESSING \
      puk_spinlock_release(); \
    } else { \
      const SYM##_func_t PUK_UNUSED _fun = (SYM##_func_t)(PUK_SYM_VIRTUAL(SYM)); \
      PROCESSING \
    } \
    return _result; \
  }
#define _PUKABI_MK_SYM_WRAPV1(SYM, RETTYPE, ARGSDECL, ARG1) \
  RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL { \
    va_list _ap; \
    va_start(_ap, ARG1); \
    long _a1 = va_arg(_ap, long); \
    long _a2 = va_arg(_ap, long); \
    va_end(_ap); \
    const SYM ## _func_t fun = (SYM ## _func_t)(PUK_SYM_REAL(SYM)); \
    puk_spinlock_acquire(); \
    RETTYPE _result = (*fun)(ARG1, _a1, _a2); \
    PUK_ABI_ERRNO_INHERIT; \
    puk_spinlock_release(); \
    return _result; }
#define _PUKABI_MK_SYM_WRAPV2(SYM, RETTYPE, ARGSDECL, ARG1, ARG2) \
  RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL { \
    va_list _ap; \
    va_start(_ap, ARG2); \
    long _a1 = va_arg(_ap, long); \
    long _a2 = va_arg(_ap, long); \
    va_end(_ap); \
    const SYM ## _func_t fun = (SYM ## _func_t)(PUK_SYM_REAL(SYM)); \
    puk_spinlock_acquire(); \
    RETTYPE _result = (*fun)(ARG1, ARG2, _a1, _a2); \
    PUK_ABI_ERRNO_INHERIT; \
    puk_spinlock_release(); \
    return _result; }
#define _PUKABI_MK_SYM_WRAPV3(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, ARG3)	\
  RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL { \
    va_list _ap; \
    va_start(_ap, ARG3); \
    long _a1 = va_arg(_ap, long); \
    long _a2 = va_arg(_ap, long); \
    va_end(_ap); \
    const SYM ## _func_t fun = (SYM ## _func_t)(PUK_SYM_REAL(SYM)); \
    puk_spinlock_acquire(); \
    RETTYPE _result = (*fun)(ARG1, ARG2, ARG3, _a1, _a2);	\
    PUK_ABI_ERRNO_INHERIT; \
    puk_spinlock_release(); \
    return _result; }

/** Generates symbol stub- 
 *  procedure version */
#define _PUKABI_MK_SYM_STUBP(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, USEREAL) \
  RETTYPE PUK_SYM_STUB(SYM) ARGSDECL { \
    if(USEREAL){ \
      const SYM##_func_t fun = (SYM##_func_t)(PUK_SYM_REAL(SYM)); \
      puk_spinlock_acquire(); \
      PREPROCESSING \
      ((*fun) ARGSUSE); \
      PUK_ABI_ERRNO_INHERIT; \
      POSTPROCESSING \
      puk_spinlock_release(); \
    } else { \
      const SYM##_func_t fun = (SYM##_func_t)(PUK_SYM_VIRTUAL(SYM)); \
      PREPROCESSING \
      ((*fun) ARGSUSE); \
      POSTPROCESSING \
    } \
  }

#define _PUKABI_MK_SYM_WRAPP(SYM, RETTYPE, ARGSDECL, ARGSUSE) \
  RETTYPE PUK_SYM_WRAP(SYM) ARGSDECL { \
  const SYM ## _func_t fun = (SYM ## _func_t)(PUK_SYM_REAL(SYM)); \
  puk_spinlock_acquire(); \
  ((*fun) ARGSUSE); \
  PUK_ABI_ERRNO_INHERIT; \
  puk_spinlock_release(); }


/** Declares a symbol descriptor structure and forward 
 *  declaration of stub and wrapper. */
#define _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  struct puk_symbol_s PUK_SYM_DESCNAME(SYM) = \
    { .ptr_real = NULL, .ptr_virtual = NULL, .name =_TOSTRING(SYM) };
#define _PUKABI_MK_SYM_DESCRIPTOR_INIT(SYM, INIT)  \
  struct puk_symbol_s PUK_SYM_DESCNAME(SYM) = \
    { .ptr_real = INIT, .ptr_virtual = NULL, .name =_TOSTRING(SYM) };

/** Creates a symbol 'foo' as an alias for '__puk_abi_stub__foo'
 */
#ifdef PUK_ABI_WRAP_INTERNALS
#  define _PUKABI_MK_SYM_ALIAS(SYM) \
    _MK_ALIAS(PUK_SYM_STUB(SYM), SYM) \
    _MK_ALIAS(PUK_SYM_STUB(SYM), _PASTE(__, SYM))
#else /* PUK_ABI_WRAP_INTERNALS */
/* intercept only public libc symbols */
#  define _PUKABI_MK_SYM_ALIAS(SYM) \
    _MK_ALIAS(PUK_SYM_STUB(SYM), SYM)
#endif /* PUK_ABI_WRAP_INTERNALS */


/** creates a type, descriptor, stub, wrapper and alias for a symbol- "function" flavor.
 */
#define MKSYMFX(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, USEREAL) \
  PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL); \
  _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  _PUKABI_MK_SYM_STUBF(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, USEREAL) \
  _PUKABI_MK_SYM_WRAPF(SYM, RETTYPE, ARGSDECL, ARGSUSE) \
  _PUKABI_MK_SYM_ALIAS(SYM)

/** creates a type, descriptor, stub, wrapper and alias for a symbol- "procedure" flavor.
 */
#define MKSYMPX(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, USEREAL) \
  PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL); \
  _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  _PUKABI_MK_SYM_STUBP(SYM, RETTYPE, ARGSDECL, ARGSUSE, PREPROCESSING, POSTPROCESSING, USEREAL) \
  _PUKABI_MK_SYM_WRAPP(SYM, RETTYPE, ARGSDECL, ARGSUSE) \
  _PUKABI_MK_SYM_ALIAS(SYM)

/** creates a type, descriptor, stub, wrapper and alias for a symbol- "varargs with 1 fixed arg" flavor.
 */
#define MKSYMVX1(SYM, RETTYPE, ARGSDECL, ARG1, PROCESSING, USEREAL) \
  PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL); \
  _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  _PUKABI_MK_SYM_STUBV(SYM, RETTYPE, ARGSDECL, PROCESSING, USEREAL) \
  _PUKABI_MK_SYM_WRAPV1(SYM, RETTYPE, ARGSDECL, ARG1) \
  _PUKABI_MK_SYM_ALIAS(SYM)
/** creates a type, descriptor, stub, wrapper and alias for a symbol- "varargs with 2 fixed args" flavor.
 */
#define MKSYMVX2(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, PROCESSING, USEREAL) \
  PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL); \
  _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  _PUKABI_MK_SYM_STUBV(SYM, RETTYPE, ARGSDECL, PROCESSING, USEREAL) \
  _PUKABI_MK_SYM_WRAPV2(SYM, RETTYPE, ARGSDECL, ARG1, ARG2) \
  _PUKABI_MK_SYM_ALIAS(SYM)
/** creates a type, descriptor, stub, wrapper and alias for a symbol- "varargs with 3 fixed args" flavor.
 */
#define MKSYMVX3(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, ARG3, PROCESSING, USEREAL) \
  PUK_ABI_MKTYPE(SYM, RETTYPE, ARGSDECL); \
  _PUKABI_MK_SYM_DESCRIPTOR(SYM) \
  _PUKABI_MK_SYM_STUBV(SYM, RETTYPE, ARGSDECL, PROCESSING, USEREAL) \
  _PUKABI_MK_SYM_WRAPV3(SYM, RETTYPE, ARGSDECL, ARG1, ARG2, ARG3)   \
  _PUKABI_MK_SYM_ALIAS(SYM)

/** creates a stub for an unsupported symbol.
 */
#define MKSYMNOSUPPORT(SYM) \
  void SYM(void) { puk_abi_fatal("Puk-ABI: symbol '%s' not supported.\n", #SYM); }


/* All standard interception symbols are generated *here*. 
 */
#include "Puk-ABI-symbols.h"
#include "Puk-ABI-alias.h"


/* ** additional symbols for malloc & co. */

#ifndef PUK_ABI_WRAP_INTERNALS
/* for some memory operations- always intercept*/
  _MK_ALIAS(PUK_SYM_STUB(malloc),  __malloc);
  _MK_ALIAS(PUK_SYM_STUB(free),    __free);
  _MK_ALIAS(PUK_SYM_STUB(free),    __cfree);
  _MK_ALIAS(PUK_SYM_STUB(realloc), __realloc);
  _MK_ALIAS(PUK_SYM_STUB(calloc),  __calloc);
#endif /* PUK_ABI_WRAP_INTERNALS */
/* for backward compatibility, defines 'cfree' as an alias for 'free' */
_MK_ALIAS(free, cfree);

/* *** errno ***********************************************
 *
 * 'puk_abi_errno_sym': #defined to the real __errno_location()
 * 'puk_abi_errno_stub': alias to use PUK_SYM_STUB(errno) from outside PukABI
 * 'puk_abi_errno_real': the *real* __errno_location(), supposedly not usefull for the end-user.
 */

#ifndef puk_abi_errno_sym
#  error "puk_abi_errno_sym not defined."
#endif

static int puk_abi_errno_use_virtual = 0;
static int puk_abi_errno_value = 0;
static int*puk_abi_errno_base_handler(void)
{
  return &puk_abi_errno_value;
}

typedef int*(*puk_abi_errno_fun_t)(void);
extern int* PUK_SYM_STUB(puk_abi_errno_sym)(void); /* forward declaration for some picky compilers */
_MK_ALIAS(PUK_SYM_STUB(puk_abi_errno_sym), puk_abi_errno_stub);
_MK_ALIAS(PUK_SYM_STUB(puk_abi_errno_sym), PUK_SYM_WRAP(puk_abi_errno_sym));
_PUKABI_MK_SYM_DESCRIPTOR_INIT(puk_abi_errno_sym, &puk_abi_errno_base_handler); 
_PUKABI_MK_SYM_ALIAS(puk_abi_errno_sym);


extern int* PUK_SYM_STUB(puk_abi_errno_sym)(void)
{
  int*result;
  if(puk_abi_errno_use_virtual)
    {
      result = (*((PUK_SYM_FUNCTYPE(puk_abi_errno_sym))
		  (PUK_SYM_VIRTUAL(puk_abi_errno_sym))))();
    }
  else
    {
      const puk_abi_errno_fun_t _fun = (puk_abi_errno_fun_t)(PUK_SYM_REAL(puk_abi_errno_sym));
      result = (*_fun)();
    }
  return result;
}

extern void puk_abi_set_errno_handler(int*(*errno_handler)(void))
{
  PUK_SYM_DESCNAME(puk_abi_errno_sym).ptr_virtual = errno_handler;
  puk_abi_errno_use_virtual = 1;
  *(puk_abi_errno_stub()) = 0; 
}

