/** @file
 */

#include "Puk-internals.h"
#include "Module.h"
#include <stdlib.h>

PADICO_MODULE_HOOK(Puk);


/** PadicoTM components manager 
 */
static struct
{
  void (*comp_acquire)(void);
  void (*comp_release)(void);
  puk_hashtable_t     ifaces;  /**< table of known interfaces- hashed by name */
  puk_hashtable_t   adapters;  /**< table of registered adapters- hashed by name */
  const char attr_undefined[];
} puk_components = { .attr_undefined = "undefined", .comp_acquire = NULL, .comp_release = NULL };


/* ********************************************************* */

static void puk_composite_start_handler(puk_parse_entity_t e);
static void puk_composite_end_handler(puk_parse_entity_t e);
static void puk_component_start_handler(puk_parse_entity_t e);
static void puk_component_end_handler(puk_parse_entity_t e);
static void puk_attr_end_handler(puk_parse_entity_t e);
static void puk_uses_start_handler(puk_parse_entity_t e);
static void puk_entrypoint_start_handler(puk_parse_entity_t e);

/* ********************************************************* */
/* local shortcuts */

/** get a receptacle of a component connection vector */
static inline puk_component_conn_t puk_component_conn_lookup(puk_component_conn_vect_t conns,
							     puk_iface_t iface, const char*label)
{
  puk_component_conn_itor_t i;
  puk_vect_foreach(i, puk_component_conn, conns)
    {
      if((i->receptacle->iface == iface) && 
	 ( (label == NULL) || 
	   (strcmp(i->receptacle->label, label) == 0) ||
	   (strcmp(i->receptacle->iface->name, label) == 0)))
	return i;
    }
  return NULL;
}

/** @internal Gets the composite content from an adapter (supposed to be a composite) */
static inline
struct puk_composite_content_s*puk_composite_get(puk_component_t adapter)
{
  const struct puk_composite_driver_s*const driver = adapter ?
    puk_adapter_get_driver_PadicoComposite(adapter, NULL) : NULL;
  if(!driver)
    {
      padico_fatal("PadicoTM: adapter *%s* has no 'PadicoComposite' facet.\n", adapter->name);
    }
  return driver->content;
}

void padico_tm_comp_set_handlers(void(*_acquire)(void), void(*_release)(void))
{
  puk_components.comp_acquire = _acquire;
  puk_components.comp_release = _release;
}

static inline void puk_components_acquire(void)
{
  if(puk_components.comp_acquire)
    (*puk_components.comp_acquire)();
}
static inline void puk_components_release(void)
{
  if(puk_components.comp_release)
    (*puk_components.comp_release)();
}

/* ********************************************************* */

void padico_puk_components_init(void)
{
  puk_components.ifaces   = puk_hashtable_new(&puk_hash_string, &puk_hash_string_eq);
  puk_components.adapters = puk_hashtable_new(&puk_hash_string, &puk_hash_string_eq);
  puk_iface_register("PadicoAdapter");
  puk_iface_register("PadicoComposite");

  /* parsing */
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "puk:composite", /* puk:composite */
    .start_handler  = &puk_composite_start_handler,
    .end_handler    = &puk_composite_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "puk:component", /* puk:component */
    .start_handler  = &puk_component_start_handler,
    .end_handler    = &puk_component_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "puk:attr", /* puk:attr */
    .start_handler  = NULL,
    .end_handler    = &puk_attr_end_handler,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "puk:uses", /* puk:uses */
    .start_handler  = &puk_uses_start_handler,
    .end_handler    = NULL,
    .required_level = PUK_TRUST_CONTROL
  });
  puk_xml_add_action((struct puk_tag_action_s)
  {
    .xml_tag        = "puk:entry-point", /* puk:entry-point */
    .start_handler  = &puk_entrypoint_start_handler,
    .end_handler    = NULL,
    .required_level = PUK_TRUST_CONTROL
  });
}

void padico_puk_components_finalize(void)
{
  /* purge adapters */
  puk_component_t c;
  puk_hashtable_getentry(puk_components.adapters, NULL, &c);
  while(c)
    {
      puk_component_destroy(c);
      puk_hashtable_getentry(puk_components.adapters, NULL, &c);
    }
  puk_hashtable_delete(puk_components.adapters);
  /* purge ifaces */
  puk_iface_t i;
  puk_hashtable_getentry(puk_components.ifaces, NULL, &i);
  while(i)
    {
      puk_iface_destroy(i);
      puk_hashtable_getentry(puk_components.ifaces, NULL, &i);
    }
  puk_hashtable_delete(puk_components.ifaces);
}


/* ********************************************************* */
/* *** PadicoComponent interface *************************** */

puk_iface_t puk_iface_register(const char*name)
{
  puk_components_acquire();
  puk_iface_t iface = puk_hashtable_lookup(puk_components.ifaces, name);
  if(!iface)
    {
      iface = padico_malloc(sizeof(struct puk_iface_s));
      padico_out(40, "registering iface %s\n", name);
      iface->name = padico_strdup(name);
      puk_hashtable_insert(puk_components.ifaces, iface->name, iface);
    }
  puk_components_release();
  return iface;
}

puk_iface_t puk_iface_lookup(const char*name)
{
  puk_components_acquire();
  puk_iface_t iface = puk_hashtable_lookup(puk_components.ifaces, name);
  if(!iface)
    {
      padico_out(6, "PadicoComponent: interface %s not found.\n", name);
    }
  puk_components_release();
  return iface;
}

void puk_iface_destroy(puk_iface_t iface)
{
  puk_hashtable_remove(puk_components.ifaces, iface->name);
  padico_free((char*)iface->name);
  padico_free(iface);
}

/* ********************************************************* */
/* ***** PadicoComponent adapter *************************** */

static puk_component_t puk_component_new(const char*name)
{
  puk_component_t c = padico_malloc(sizeof(struct puk_adapter_s));
  padico_out(40, "building component %s\n", name);
  c->name = padico_strdup(name);
  c->attrs = NULL;
  puk_uses_vect_init(&c->uses);
  puk_facet_vect_init(&c->provides);
  return c;
}

puk_component_t puk_component_declare2(const char*adapter_name, int n_ports,
				       const struct puk_component_port_s ports[])
{
  int i;
  puk_components_acquire();
  puk_component_t a = puk_hashtable_lookup(puk_components.adapters, adapter_name);
  puk_components_release();
  if(a != NULL)
    return a;
  a = puk_component_new(adapter_name);
  for(i = 0; i < n_ports; i++)
    {
      struct puk_component_port_s r = ports[i];
      switch(r.kind)
	{
	case PUK_COMPONENT_PROVIDES:
	  {
	    puk_iface_t iface = puk_iface_register(r.content.provides.iface_name);
	    padico_out(40, "  adapter provides facet %s\n", iface->name);
	    const struct puk_facet_s _facet =
	      {
		.iface  = iface,
		.label  = r.label?padico_strdup(r.label):NULL,
		.driver = r.content.provides.driver
	      };
	    puk_facet_vect_push_back(&a->provides, _facet);
	  }
 	  break;
	case PUK_COMPONENT_USES:
	  {
	    puk_iface_t iface = puk_iface_register(r.content.uses.iface_name);
	    padico_out(40, "  adapter uses iface %s\n", iface->name);
	    const struct puk_uses_s _uses =
	      {
		.iface = iface,
		.label = r.label
	      };
	    puk_uses_vect_push_back(&a->uses, _uses);
	  }
	  break;
	case PUK_COMPONENT_ATTR:
	  {
	    const char*label = r.label;
	    const char*value = r.content.attr.value?:puk_components.attr_undefined;
	    if(!a->attrs)
	      a->attrs = puk_hashtable_new_string();
	    puk_hashtable_insert(a->attrs, label, (void*)value);
	  }
	  break;
	}
    }
  padico_out(40, "registering adapter %s.\n", a->name);
  puk_components_acquire();
  puk_hashtable_insert(puk_components.adapters, a->name, a);
  puk_components_release();
  return a;
}

puk_facet_t puk_adapter_get_facet(puk_component_t a, puk_iface_t iface, const char*label)
{
  puk_facet_itor_t i;
  /* @note linear search better than anything else when #facet < 5! */
  puk_vect_foreach(i, puk_facet, &a->provides)
    {
      if(i->iface == iface && 
	 ( (!label) ||
	   ((label != NULL) && (i->label != NULL) && strcmp(label, i->label) == 0)))
	return i;
    }
  return NULL;
}

/** Adapter lookup+auto-load if lookup fails.
 */
puk_component_t puk_adapter_resolve(const char*adapter_name)
{
  puk_components_acquire();
  puk_component_t adapter = puk_hashtable_lookup(puk_components.adapters, adapter_name);
  puk_components_release();
  if(!adapter)
    {
      padico_mod_t mod = NULL;
      padico_rc_t rc = padico_puk_mod_resolve(&mod, adapter_name);
      padico_out(10, "component %s not found. Auto-load...\n", adapter_name);
      if(!padico_rc_iserror(rc))
	{
	  padico_out(6, "loading component %s ...\n", adapter_name);
	  puk_mod_set_ancestor(mod, padico_module_self());
	  rc = padico_puk_mod_load(mod);
	  puk_components_acquire();
	  adapter = puk_hashtable_lookup(puk_components.adapters, adapter_name);
	  puk_components_release();
	  assert(adapter != NULL);
	}
      if(!padico_rc_iserror(rc))
	{
	  padico_out(10, "component %s successfully loaded.\n", adapter_name);
	}
      else
	{
	  padico_warning("PadicoTM: component %s auto-load failed.\n", adapter_name);
	}
    }
  return adapter;
}

void puk_component_destroy(puk_component_t c)
{
  puk_hashtable_remove(puk_components.adapters, c->name);
  const struct puk_composite_driver_s*const composite_driver = 
    puk_adapter_get_driver_PadicoComposite(c, NULL);
  if(composite_driver)
    {
      struct puk_composite_content_s*content = composite_driver->content; 
      puk_component_conn_vect_itor_t ep;
      for(ep  = puk_component_conn_vect_begin(&content->entry_points);
	  ep != puk_component_conn_vect_end(&content->entry_points);
	  ep  = puk_component_conn_vect_next(ep))
	{
	  struct puk_uses_s*receptacle = ep->receptacle;
	  padico_free((char*)receptacle->label);
	  padico_free(receptacle);
	}
      puk_component_conn_vect_destroy(&content->entry_points);
      puk_context_t context;
      while(puk_hashtable_getentry(content->contexts, NULL, &context), (context != NULL))
	{
	  puk_hashtable_remove(content->contexts, context->id);
	  if(context->attrs)
	    {
	      char*label, *value;
	      while(puk_hashtable_getentry(context->attrs, &label, &value), (label != NULL))
		{
		  puk_hashtable_remove(context->attrs, label);
		  padico_free(label);
		  padico_free(value);
		}
	      puk_hashtable_delete(context->attrs);
	    }
	  padico_free((char*)context->id);
	  padico_free(context);
	}
      puk_hashtable_delete(content->contexts);
      padico_free(content);
    }
  padico_free((char*)c->name);
  puk_uses_vect_destroy(&c->uses);
  puk_facet_vect_itor_t f;
  for(f  = puk_facet_vect_begin(&c->provides);
      f != puk_facet_vect_end(&c->provides);
      f  = puk_facet_vect_next(f))
    {
      if(f->label)
	padico_free((char*)f->label);
    }
  puk_facet_vect_destroy(&c->provides);
  if(c->attrs)
    {
      puk_hashtable_delete(c->attrs);
    }
  padico_free(c);

}

/* ********************************************************* */
/* *** Adapter context ************************************* */


void puk_context_putattr(puk_context_t context, 
			 const char*label, const char*value)
{
  if(!context->attrs)
    context->attrs = puk_hashtable_new_string();
  puk_hashtable_insert(context->attrs, padico_strdup(label), padico_strdup(value));
}

const char*puk_context_getattr(puk_context_t context,
			       const char*label)
{
  const char*value = context->attrs?puk_hashtable_lookup(context->attrs, label):NULL;
  return value;
}

void puk_context_indirect(puk_instance_t instance,
			  puk_iface_t iface,
			  const char*label,
			  void*_driver, void *_status)
{
  const puk_component_conn_t entry_point = puk_component_conn_lookup(&instance->context->conns, iface, label);
  assert(entry_point != NULL);
  if(_driver)
    {
      *(const void**)_driver = entry_point->facet->driver;
    }
  if(_status)
    {
      assert(puk_composite_get(instance->container->adapter) != NULL);
      const struct puk_composite_status_s*composite_status = instance->container->status;
      puk_instance_t sub_instance = puk_hashtable_lookup(composite_status->instance_table, entry_point->context);
      puk_instance_entry_point(sub_instance, iface, label, NULL, _status);
    }
}


/* ********************************************************* */
/* *** Composite ******************************************* */

static void*puk_composite_instanciate(puk_instance_t instance, puk_context_t in_context)
{
  struct puk_composite_status_s*status = padico_malloc(sizeof(struct puk_composite_status_s));
  instance->status = status;
  if(in_context)
    {
      padico_warning("PadicoTM: instances inside an instance.\n");
    }
  status->instance_table = puk_hashtable_new_ptr();
  status->instance = instance;
  puk_context_t context = NULL;
  puk_adapter_enumerator_t e = puk_adapter_enumerator_new(instance->adapter, BOTTOM_UP);
  for(context  = puk_adapter_enumerator_next(e);
      context != NULL;
      context  = puk_adapter_enumerator_next(e))
    {
      puk_instance_t sub_instance = puk_component_instanciate(context->adapter, instance, context);
      puk_hashtable_insert(status->instance_table, context, sub_instance);
    }
  puk_adapter_enumerator_delete(e);
  return status;
}

static void puk_composite_destroy(void*_status)
{
  struct puk_composite_status_s*composite_status = _status;
  puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(composite_status->instance_table);
  puk_instance_t sub_instance = puk_hashtable_enumerator_next_data(e);
  while(sub_instance)
    {
      puk_instance_destroy(sub_instance);
      sub_instance = puk_hashtable_enumerator_next_data(e);
    }
  puk_hashtable_enumerator_delete(e);
  puk_hashtable_delete(composite_status->instance_table);
  padico_free(composite_status);
}

static struct puk_adapter_driver_s composite_adapter_driver =
  {
    .instanciate = &puk_composite_instanciate,
    .destroy     = &puk_composite_destroy
  };


/* ********************************************************* */
/* *** Adapter instance ************************************ */

puk_instance_t puk_component_instanciate(puk_component_t adapter, puk_instance_t container, puk_context_t context)
{
  const struct puk_adapter_driver_s*driver = puk_adapter_get_driver_PadicoAdapter(adapter, NULL);
  puk_instance_t instance = padico_malloc(sizeof(struct puk_instance_s));
  instance->adapter   = adapter;
  instance->status    = NULL;
  instance->container = container;
  instance->context   = context;
  if(driver && driver->instanciate)
    {
      instance->status = (*driver->instanciate)(instance, context);
    }
  return instance;
}

void puk_instance_destroy(puk_instance_t instance)
{
  if(!instance)
    return;

  const struct puk_adapter_driver_s*driver =
    puk_adapter_get_driver_PadicoAdapter(instance->adapter, NULL);
  if(driver && driver->destroy)
    {
      (*driver->destroy)(instance->status);
    }
  padico_free(instance);
}

void puk_instance_entry_point(puk_instance_t instance,
			      puk_iface_t iface,
			      const char*label,
			      void*_driver, void*_status)
{
  const puk_component_t adapter = instance->adapter;
  if(_driver)
    {
      puk_facet_t f = puk_adapter_get_facet(adapter, iface, label);
      *((const void**)_driver) = f->driver;
    }
  if(_status)
    {
      const struct puk_composite_driver_s*composite_driver =
	puk_adapter_get_driver_PadicoComposite(adapter, NULL);
      if(composite_driver)
	{
	  struct puk_composite_content_s*content = composite_driver->content;
	  const struct puk_composite_status_s*composite_status = instance->status;
	  puk_component_conn_itor_t i;
	  puk_vect_foreach(i, puk_component_conn, &content->entry_points)
	    {
	      puk_instance_t sub_instance = 
		puk_hashtable_lookup(composite_status->instance_table, i->context);
	      puk_instance_entry_point(sub_instance, iface, label, NULL, _status);
	      if(*(void**)_status)
		break;
	    }
	}
      else
	{
	  *((void**)_status) = instance->status;
	}
    }
}

puk_instance_t puk_instance_find_context(puk_instance_t container, puk_instance_t local)
{
  if((container->adapter == local->adapter) && (container->context == local->context))
    return container;
  else
    {
      const struct puk_composite_driver_s*const composite_driver = 
	puk_adapter_get_driver_PadicoComposite(container->adapter, NULL);
      if(composite_driver)
	{
	  struct puk_composite_content_s*content = composite_driver->content;
	  const struct puk_composite_status_s*composite_status = container->status;
	  puk_component_conn_itor_t i;
	  puk_vect_foreach(i, puk_component_conn, &content->entry_points)
	    {
	      puk_instance_t sub_instance = 
		puk_hashtable_lookup(composite_status->instance_table, i->context);
	      puk_instance_t instance = puk_instance_find_context(sub_instance, local);
	      if(instance)
		return instance;
	    }
	}
    }
  return NULL;
}


/* ********************************************************* */
/* *** Assembly marshalling ******************************** */

padico_string_t puk_adapter_serialize(puk_component_t component)
{
  padico_string_t s = padico_string_new();
  if(puk_adapter_get_driver_PadicoComposite(component, NULL))
    {
      padico_string_catf(s, "<puk:composite id=\"_puk:_x(%p):%s\">\n", component, component->name);
      /* adapters */
      puk_adapter_enumerator_t adapters = puk_adapter_enumerator_new(component, BOTTOM_UP);
      puk_context_t context = puk_adapter_enumerator_next(adapters);
      puk_component_conn_vect_itor_t e;
      while(context)
	{
	  padico_string_catf(s, "  <puk:component name=\"%s\" id=\"%p\">\n", 
			     context->adapter->name, context);
	  /* receptacles (uses) */
	  puk_vect_foreach(e, puk_component_conn, &context->conns)
	    {
	      if(e->context != NULL)
		{
		  padico_string_catf(s, "    <puk:uses iface=\"%s\" port=\"%s\" provider-id=\"%p\" provider-port=\"%s\" />\n", 
				     e->receptacle->iface->name, e->receptacle->label,
				     e->context, e->facet->label);
		}
	    }
	  /* attributes */
	  if(context->attrs)
	    {
	      puk_hashtable_enumerator_t attrs = puk_hashtable_enumerator_new(context->attrs);
	      char*label, *value; 
	      puk_hashtable_enumerator_next2(attrs, &label, &value);
	      while(label)
		{
		  padico_string_catf(s, "    <puk:attr label=\"%s\">%s</puk:attr>\n", 
				     label, value);
		  puk_hashtable_enumerator_next2(attrs, &label, &value);
		}
	      puk_hashtable_enumerator_delete(attrs);
	    }
	  padico_string_catf(s, "  </puk:component>\n");
	  context = puk_adapter_enumerator_next(adapters);
	}
      puk_adapter_enumerator_delete(adapters);
      /* exported facets */
      struct puk_composite_content_s*content = puk_composite_get(component);
      puk_vect_foreach(e, puk_component_conn, &content->entry_points)
	{
	  padico_string_catf(s, "<puk:entry-point iface=\"%s\" port=\"%s\" provider-id=\"%p\" provider-port=\"%s\"/>\n",
			     e->receptacle->iface->name, e->receptacle->label, 
			     e->context, e->facet->label);
	}
      padico_string_catf(s, "</puk:composite>\n");
    }
  else
    {
      padico_string_catf(s, "  <puk:component name=\"%s\"/>\n", component->name);
    }
  return s;
}

puk_component_t puk_adapter_parse(const char*s)
{
  puk_component_t adapter = NULL;
  struct puk_parse_entity_s e = puk_xml_parse_buffer(s, strlen(s), PUK_TRUST_LOCAL);
  if(puk_parse_is(&e, "puk:composite") || puk_parse_is(&e, "puk:component"))
    {
      adapter = puk_parse_get_content(&e);
    }
  else
    {
      padico_fatal("PadicoTM: error while parsing *%s*: not a component.\n", s);
    }
  return adapter;
}

puk_component_t puk_adapter_parse_file(const char*filename)
{
  puk_component_t adapter = NULL;
  struct puk_parse_entity_s e = puk_xml_parse_file(filename, PUK_TRUST_LOCAL);
  padico_rc_t rc = puk_parse_get_rc(&e);
  if(padico_rc_iserror(rc))
    {
      padico_fatal("Puk: error while parsing file %s: *%s*.\n",
		   filename, padico_rc_gettext(rc));
    }
  else if(puk_parse_is(&e, "puk:composite") || puk_parse_is(&e, "puk:component"))
    {
      adapter = puk_parse_get_content(&e);
    }
  else
    {
      padico_fatal("PadicoTM: error while parsing file %s: not a component (%s).\n", filename, e.tag);
    }
  return adapter;
}


/* ********************************************************* */
/* *** Assembly parsing ************************************ */


/* *** <puk:composite> */
static void puk_composite_start_handler(puk_parse_entity_t e)
{
  const char*id  = puk_parse_getattr(e, "id");
  const char*ref = puk_parse_getattr(e, "ref");
  puk_component_t assembly = NULL;
  assert(!ref != !id);
  if(ref)
    {
      assembly = puk_adapter_resolve(ref);
    }
  else
    {
      assembly = puk_component_new(id);
      const struct puk_facet_s adapter_facet =
	{
	  .iface  = puk_iface_PadicoAdapter(),
	  .driver = &composite_adapter_driver,
	  .label  = NULL
	};
      puk_facet_vect_push_back(&assembly->provides, adapter_facet);
      struct puk_composite_content_s*composite_content = padico_malloc(sizeof(struct puk_composite_content_s));
      composite_content->driver.content = composite_content;
      const struct puk_facet_s composite_facet =
	{
	  .iface  = puk_iface_PadicoComposite(),
	  .driver = &composite_content->driver,
	  .label  = NULL
	};
      puk_facet_vect_push_back(&assembly->provides, composite_facet);
      puk_component_conn_vect_init(&composite_content->entry_points);
      composite_content->contexts = puk_hashtable_new_string();
      composite_content->checked = 0;
      padico_out(40, "new assembly- id=%s\n", id);
    }
  puk_parse_set_content(e, assembly);
}
static void puk_composite_end_handler(puk_parse_entity_t e)
{
  puk_component_t composite = puk_parse_get_content(e);
  if(composite)
    {
      const struct puk_composite_driver_s*const composite_driver = 
	puk_adapter_get_driver_PadicoComposite(composite, NULL);
      if(composite_driver)
	{
	  struct puk_composite_content_s*content = composite_driver->content;
	  if(puk_hashtable_size(content->contexts) == 0 ||
	     puk_component_conn_vect_size(&content->entry_points) == 0)
	    {
	      padico_warning("PadicoTM: invalid composite '%s'- skipping.\n", composite->name);
	      puk_component_destroy(composite);
	      composite = NULL;
	    }
	  else if(!puk_parse_getattr(e, "ref"))
	    {
	      puk_components_acquire();
	      const puk_component_t a = puk_hashtable_lookup(puk_components.adapters, composite->name);
	      if(a)
		{
		  padico_out(6, "duplicate composite '%s'\n", composite->name);
		  puk_component_destroy(composite);
		  composite = a;
		}
	      else
		{
		  puk_hashtable_insert(puk_components.adapters, composite->name, composite);
		}
	      puk_components_release();
	    }
	}
    }
  puk_parse_set_content(e, composite);
}

/* *** <puk:component> */
static void puk_component_start_handler(puk_parse_entity_t e)
{
  const char*adapter_name = puk_parse_getattr(e, "name");
  const char*adapter_id   = puk_parse_getattr(e, "id");
  const char*ref = puk_parse_getattr(e, "ref");
  puk_component_t adapter = NULL;
  if(ref)
    adapter = puk_adapter_resolve(ref);
  if(adapter_name)
    adapter= puk_adapter_resolve(adapter_name);
  if(adapter == NULL)
    {
      padico_warning("Puk: component '%s' not found- skipping.\n", adapter_name);
      puk_parse_set_content(e, NULL);
    }
  else
    {
      puk_component_t assembly = puk_parse_get_content(puk_parse_parent(e));
      puk_context_t context = padico_malloc(sizeof(struct puk_generic_context_s));
      assert(adapter != NULL);
      context->adapter  = adapter;
      context->assembly = assembly;
      context->attrs    = NULL;
      /* instanciate receptacles */
      puk_component_conn_vect_init(&context->conns);
      {
	puk_uses_vect_itor_t i;
	puk_vect_foreach(i, puk_uses, &adapter->uses)
	  {
	    struct puk_component_conn_s uses =
	      { .receptacle = i, .context = NULL, .facet = NULL };
	    puk_component_conn_vect_push_back(&context->conns, uses);
	  }
      }
      if(!adapter_id)
	{
	  padico_string_t sid = padico_string_new();
	  padico_string_printf(sid, "_puk:context-%s-(%p)", adapter->name, adapter);
	  context->id = padico_strdup(padico_string_get(sid));
	  padico_string_delete(sid);
	}
      else
	context->id = padico_strdup(adapter_id);
      puk_hashtable_insert(puk_composite_get(assembly)->contexts, context->id, context);


      padico_out(40, "created adapter name=%s; id=%s\n", adapter_name, adapter_id);
      puk_parse_set_content(e, context);
    }
}
static void puk_component_end_handler(puk_parse_entity_t e)
{
  puk_context_t context = puk_parse_get_content(e);
  if(context)
    {
      /* check receptacles */
      puk_component_conn_itor_t u;
      padico_rc_t rc = padico_rc_ok();
      puk_vect_foreach(u, puk_component_conn, &context->conns)
	{
	  if(!u->facet)
	    {
	      rc = padico_rc_error("Receptacle '%s' not connected in context '%s'.\n", 
				   u->receptacle->iface->name, context->adapter->name);
	      padico_warning("PadicoTM: receptacle '%s' not connected in context '%s' (%s).\n", 
			     u->receptacle->iface->name, context->id, context->adapter->name);
	    }
	}
      /* check attributes */
      if(context->adapter->attrs)
	{
	  puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(context->adapter->attrs);
	  const char*label = NULL;
	  char*value = NULL;
	  puk_hashtable_enumerator_next2(e, &label, &value);
	  while(label)
	    {
	      if((!context->attrs) || (context->attrs && !puk_hashtable_lookup(context->attrs, label)))
		{
		  if(value != puk_components.attr_undefined)
		    {
		      /* take the default value */
		      puk_context_putattr(context, label, value);
		    }
		  else
		    {
		      rc = padico_rc_error("Attribute '%s' has no value in context '%s'.\n",
					   label, context->adapter->name);
		      padico_warning("PadicoTM: attribute '%s' has no value in context '%s'.\n",
				     label, context->adapter->name);
		      break;
		    }
		}
	      puk_hashtable_enumerator_next2(e, &label, &value);
	    }
	  puk_hashtable_enumerator_delete(e);
	}
      puk_parse_set_rc(e, rc);
    }
}

/* *** <puk:attr> */
static void puk_attr_end_handler(puk_parse_entity_t e)
{
  const char*label = puk_parse_getattr(e, "label");
  const char*value = puk_parse_get_text(e);
  puk_context_t context = 
    puk_parse_get_content_checked(puk_parse_parent(e), "puk:component");
  if(context)
    {
      assert(context && label && value);
      puk_context_putattr(context, label, value);
      padico_out(40, "created attr %s\n", label);
    }
}

/* *** <puk:uses> */
static void puk_uses_start_handler(puk_parse_entity_t e)
{
  const char*iface_name    = puk_parse_getattr(e, "iface");
  puk_iface_t iface        = puk_iface_register(iface_name);
  const char*port_name     = puk_parse_getattr(e, "port");
  const char*provider_port = puk_parse_getattr(e, "provider-port");
  const char*provider_id   = puk_parse_getattr(e, "provider-id");
  if(!iface)
    padico_fatal("PadicoTM: invalid 'uses'- interface '%s' does not exist.\n", iface_name);
  puk_context_t user = puk_parse_get_content(puk_parse_parent(e));
  if(user)
    {
      const puk_component_t composite = puk_parse_get_content(puk_parse_parent(puk_parse_parent(e)));
      if(!composite)
	padico_fatal("PadicoTM: invalid 'uses'- not inside a composite.\n");
      struct puk_composite_content_s*content = puk_composite_get(composite);
      puk_context_t provider = puk_hashtable_lookup(content->contexts, provider_id);
      if(provider)
	{
	  const puk_facet_t facet = puk_adapter_get_facet(provider->adapter, iface, provider_port);
	  assert(facet != NULL);
	  puk_component_conn_t conn = puk_component_conn_lookup(&user->conns, iface, port_name);
	  assert(conn != NULL);
	  conn->context = provider;
	  conn->facet = facet;
	}
      else
	{
	  padico_warning("PadicoTM: invalid 'uses'- context '%s' does not exist.\n", provider_id);
	}
    }
}

/* *** <puk:entry-point> */
static void puk_entrypoint_start_handler(puk_parse_entity_t e)
{
  const char*iface_name    = puk_parse_getattr(e, "iface");
  puk_iface_t iface        = puk_iface_register(iface_name);
  const char*port_name     = puk_parse_getattr(e, "port");
  const char*provider_port = puk_parse_getattr(e, "provider-port");
  const char*provider_id   = puk_parse_getattr(e, "provider-id");
  if(!iface)
    padico_fatal("PadicoTM: invalid 'entry_point'- interface '%s' does not exist.\n", iface_name);
  puk_component_t composite = puk_parse_get_content(puk_parse_parent(e));
  struct puk_composite_content_s*content = puk_composite_get(composite);
  if(composite)
    {
      puk_context_t context = puk_hashtable_lookup(content->contexts, provider_id);
      if(context)
	{
	  const puk_facet_t provider_facet =  puk_adapter_get_facet(context->adapter, iface, provider_port);
	  assert(iface != NULL && provider_facet != NULL && context != NULL);
	  struct puk_uses_s*receptacle = padico_malloc(sizeof(struct puk_uses_s));
	  receptacle->iface = iface;
	  receptacle->label = padico_strdup(port_name?:iface->name);
	  const struct puk_component_conn_s entry_point =
	    {
	      .receptacle = receptacle,
	      .context    = context,
	      .facet      = provider_facet
	    };
	  puk_component_conn_vect_push_back(&content->entry_points, entry_point);
	  const struct puk_facet_s facet =
	    {
	      .iface  = iface,
	      .driver = provider_facet->driver,
	      .label  = padico_strdup(port_name?:iface->name)
	    };
	  puk_facet_vect_push_back(&composite->provides, facet);
	}
    }
}


/* ********************************************************* */
/* *** Enumerator ****************************************** */

struct puk_adapter_enumerator_s
{
  puk_context_vect_t adapters;
  puk_context_itor_t current;
  enum adapter_enumerator_way_e way;
};

puk_context_t puk_adapter_enumerator_next(puk_adapter_enumerator_t e)
{
  puk_context_t a = NULL;
  switch(e->way)
    {
    case BOTTOM_UP:
      if(e->current != puk_context_vect_end(e->adapters))
	{
	  a = *e->current;
	  e->current = puk_context_vect_next(e->current);
	}
      else
	{
	  e->current = NULL;
	  a = NULL;
	}
      break;
    case TOP_DOWN:
      if(e->current != puk_context_vect_rend(e->adapters))
	{
	  a = *e->current;
	  e->current = puk_context_vect_rnext(e->current);
	}
      else
	{
	  e->current = NULL;
	  a = NULL;
	}
      break;
    }
  return a;
}

void puk_adapter_enumerator_delete(puk_adapter_enumerator_t e)
{
  puk_context_vect_delete(e->adapters);
  padico_free(e);
}

puk_adapter_enumerator_t
puk_adapter_enumerator_new(puk_component_t assembly, int way)
{
  puk_adapter_enumerator_t e = padico_malloc(sizeof(struct puk_adapter_enumerator_s));
  e->way      = way;
  e->adapters = puk_context_vect_new();
  e->current  = NULL;
  /* convert hashtable into unsorted vector
   */
  struct puk_composite_content_s*content = puk_composite_get(assembly);
  puk_hashtable_enumerator_t h = puk_hashtable_enumerator_new(content->contexts);
  puk_context_vect_t unsorted = puk_context_vect_new();
  puk_context_t c = puk_hashtable_enumerator_next_data(h);
  while(c)
    {
      assert(c->assembly == assembly);
      padico_out(60, "adding context id=%s; adapter=%s; assembly=%s\n", c->id, c->adapter->name, c->assembly->name);
      puk_context_vect_push_back(unsorted, c);
      c = puk_hashtable_enumerator_next_data(h);
    }
  puk_hashtable_enumerator_delete(h);
  /* topological sorting based on zero-in method (strictly speaking,
   * this is implemented as zero-out rather than zero-in) */
  const int degree = puk_context_vect_size(unsorted);
  int*matrix = padico_calloc(degree*degree, sizeof(int));
  int*done   = padico_calloc(degree, sizeof(int));
  /* matrix[x*degree + y] == 1 ~~ "x uses y" */
  int i, j, k;
  puk_context_t a_i, a_j;
  assert(degree > 0);
  /* convert the assembly into an adjacency matrix */
  for(i = 0; i < degree; i++)
    {
      a_i = puk_context_vect_at(unsorted, i);
      padico_out(60, "scanning dependancies of context i=%s\n", a_i->id);
      for(k = 0; k < puk_component_conn_vect_size(&a_i->conns); k++)
	{
	  for(j = 0; j < degree; j++)
	    {
	      a_j = puk_context_vect_at(unsorted, j);
	      if(puk_component_conn_vect_ptr(&a_i->conns, k)->context == a_j)
		{
		  padico_out(60, "context i=%s uses context j=%s\n", a_i->id, a_j->id);
		  matrix[i*degree+j] = 1;
		}
	    }
	}
    }
  for(k = 0; k < degree; k++)
    {
      for(i = 0; i < degree; i++)
	{
	  /* find a "zero-out" adapter- trying context 'i' */
	  padico_out(50, "searching for zero-out- k=%d; i=%d\n", k, i);
	  if(done[i])
	    {
	      padico_out(50, "i=%d is already done- skipping.\n", i);
	      goto next_i;
	    }
	  for(j = 0; j < degree; j++)
	    {
	      if(matrix[i*degree+j])
		{
		  padico_out(50, "i=%d is not zero-out- has out for j=%d\n", i, j);
		  goto next_i;
		}
	    }
	  padico_out(50, "i=%d is zero-out- clearing dependancies\n", i);
	  done[i] = 1;
	  /* clear all dependancies to context 'i' */
	  puk_context_t context = puk_context_vect_at(unsorted, i);
	  puk_context_vect_push_back(e->adapters, context);
	  for(j = 0; j < degree; j++)
	    {
	      matrix[j*degree+i] = 0;
	    }		  
	  goto next_k;
	next_i:
	  continue;
	}
    next_k:
      continue;
    }
  padico_free(matrix);
  padico_free(done);
  puk_context_vect_delete(unsorted);
  switch(way)
    {
    case BOTTOM_UP:
      e->current = puk_context_vect_begin(e->adapters);
      break;
    case TOP_DOWN:
      e->current = puk_context_vect_rbegin(e->adapters);
      break;
    }
  return e;
}


static
void puk_adapter_check_one(puk_component_t assembly)
{
  const struct puk_composite_driver_s*const driver = 
    puk_adapter_get_driver_PadicoComposite(assembly, NULL);
  if(driver)
    {  
      puk_adapter_enumerator_t e = puk_adapter_enumerator_new(assembly, TOP_DOWN);
      puk_adapter_enumerator_delete(e);
    }
}

void puk_adapter_check(puk_component_t assembly)
{
  if(assembly)
    {
      const struct puk_composite_driver_s*const driver =
	puk_adapter_get_driver_PadicoComposite(assembly, NULL);
      padico_print("PadicoTM: TMP- check assembly %p '%s' driver=%p ...\n", assembly, assembly->name, driver);
      puk_adapter_check_one(assembly);
      padico_print("PadicoTM: TMP- check assembly %p '%s'- done.\n", assembly, assembly->name);
    }
  puk_hashtable_enumerator_t h = puk_hashtable_enumerator_new(puk_components.adapters);
  puk_component_t a = puk_hashtable_enumerator_next_data(h);
  while(a)
    {
      puk_adapter_check_one(a);
      a = puk_hashtable_enumerator_next_data(h);
    }
}
